/* sysdep_poll.c - poll(2) compliant mainloop
 *
 * Copyright (C) 2007 Timo Teräs <timo.teras@iki.fi>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 or later as
 * published by the Free Software Foundation.
 *
 * See http://www.gnu.org/ for details.
 */

#include <poll.h>
#include "nhrp_defines.h"
#include "nhrp_common.h"

#define MAX_FDS 8

struct pollctx {
	int (*callback)(void *ctx, int fd, short events);
	void *ctx;
};

int nhrp_running = 0;
struct nhrp_task_list nhrp_all_tasks;

static int numfds = 0;
static struct pollfd gfds[MAX_FDS];
static struct pollctx gctx[MAX_FDS];

int nhrp_task_poll_fd(int fd, short events, int (*callback)(void *ctx, int fd, short events),
		      void *ctx)
{
	if (numfds >= MAX_FDS) {
		nhrp_error("Poll table full. Increase MAX_FDS in sysdep_poll.c.");
		return FALSE;
	}

	gctx[numfds].callback = callback;
	gctx[numfds].ctx = ctx;
	gfds[numfds].fd = fd;
	gfds[numfds].events = events;
	numfds++;

	return TRUE;
}

static void nhrp_task_unpoll_index(int i)
{
	gfds[i] = gfds[numfds - 1];
	gctx[i] = gctx[numfds - 1];
	numfds--;
}

void nhrp_task_unpoll_fd(int fd)
{
	int i;

	for (i = 0; i < numfds; i++)
		if (gfds[i].fd == fd)
			break;

	if (i < numfds)
		nhrp_task_unpoll_index(i);
}

void nhrp_task_schedule(struct nhrp_task *task, int timeout, const struct nhrp_task_ops *ops)
{
	struct nhrp_task *after = NULL, *next;

	nhrp_task_cancel(task);

	gettimeofday(&task->execute_time, NULL);
	task->ops = ops;
	task->execute_time.tv_usec += (timeout % 1000) * 1000;
	task->execute_time.tv_sec += timeout / 1000 +
		(task->execute_time.tv_usec / 1000000);
	task->execute_time.tv_usec %= 1000000;

	for (next = LIST_FIRST(&nhrp_all_tasks);
	     next != NULL && timercmp(&task->execute_time, &next->execute_time, >);
	     next = LIST_NEXT(next, task_list))
		after = next;

	if (after != NULL)
		LIST_INSERT_AFTER(after, task, task_list);
	else
		LIST_INSERT_HEAD(&nhrp_all_tasks, task, task_list);
}

void nhrp_task_cancel(struct nhrp_task *task)
{
	if (task->ops != NULL) {
		LIST_REMOVE(task, task_list);
		task->ops = NULL;
	}
}

void nhrp_task_run(void)
{
	struct timeval now;
	struct nhrp_task *task;
	int i, timeout;

	nhrp_running = TRUE;
	do {
		if (numfds == 0 && LIST_EMPTY(&nhrp_all_tasks))
			break;

		gettimeofday(&now, NULL);
		while (!LIST_EMPTY(&nhrp_all_tasks) && timercmp(&LIST_FIRST(&nhrp_all_tasks)->execute_time, &now, <=)) {
			const struct nhrp_task_ops *ops;

			task = LIST_FIRST(&nhrp_all_tasks);
			ops = task->ops;

			nhrp_task_cancel(task);
			ops->callback(task);
		}

		if (!LIST_EMPTY(&nhrp_all_tasks)) {
			task = LIST_FIRST(&nhrp_all_tasks);

			timeout = task->execute_time.tv_sec - now.tv_sec;
			timeout *= 1000;
			timeout += (task->execute_time.tv_usec - now.tv_usec) / 1000;
		} else {
			timeout = -1;
		}

		poll(gfds, numfds, timeout);

		for (i = 0; i < numfds; i++) {
			if (gfds[i].revents) {
				if (gctx[i].callback(gctx[i].ctx, gfds[i].fd,
						     gfds[i].revents))
					nhrp_task_unpoll_index(i);
			}
		}
	} while (nhrp_running);
}

void nhrp_task_stop(void)
{
	nhrp_running = FALSE;
}
